from ac_segmentation.neurotorch.core.trainer import TrainerDecorator
import tensorboardX
import os
import logging
import time
import numpy as np


class LossWriter(TrainerDecorator):
    """
    Logs the loss at each iteration to a Tensorboard log
    """
    def __init__(self, trainer, logger_dir, experiment_name):
        """
        Initializes the Tensorboard writer

        :param trainer: Trainer object that the class wraps
        :param logger_dir: Directory to save Tensorboard logs
        :param experiment_name: The name to mark the experiment
        """
        if not os.path.isdir(logger_dir):
            raise IOError("{} is not a valid directory".format(logger_dir))

        super().__init__(trainer)
        experiment_dir = os.path.join(logger_dir, experiment_name)
        os.makedirs(experiment_dir, exist_ok=True)
        self.train_writer = tensorboardX.SummaryWriter(os.path.join(experiment_dir,
                                                       "train_log"))
        self.validation_writer = tensorboardX.SummaryWriter(os.path.join(experiment_dir,
                                                            "validation_log"))

        self.iteration = 0

    def log_loss(self, loss: float, duration: float, iteration: int):
        """
        Writes the loss onto the Tensorboard log

        :param loss: The training loss of the model
        :param duration: The time elapsed by the current iteration
        :param iteration: The current iteration of the model
        """
        self.train_writer.add_scalar("Time", duration, iteration)
        self.train_writer.add_scalar("Loss", loss, iteration)

    def evaluate(self, batch):
        start = time.time()
        loss, accuracy, output = super().evaluate(batch)
        end = time.time()

        duration = end - start

        self.validation_writer.add_scalar("Time", duration, self.iteration)
        self.validation_writer.add_scalar("Loss", loss, self.iteration)
        self.validation_writer.add_scalar("Accuracy", accuracy*100, self.iteration)

        return loss, accuracy, output

    def run_epoch(self, sample_batch):
        """
        Runs an epoch and saves the parameters onto the Tensorboard log

        :param sample_batch: A batch of input/label samples for training
        """
        start = time.time()
        loss = super().run_epoch(sample_batch)
        end = time.time()

        duration = end - start

        self.iteration += 1

        if self.iteration % 10 == 0:
            self.log_loss(loss, duration, self.iteration)

        return loss


class TrainingLogger(TrainerDecorator):
    """
    Logs the iteration parameters onto a plain text log file
    """
    def __init__(self, trainer, logger_dir=None):
        """
        Initializes the Python Logger

        :param trainer: Trainer object that the class wraps
        :param logger_dir: The directory to save logs
        """
        if logger_dir is not None and not os.path.isdir(logger_dir):
            raise IOError("{} is not a valid directory".format(logger_dir))

        super().__init__(trainer)

        self.logger = logging.getLogger("Trainer")
        self.logger.setLevel(logging.INFO)

        if logger_dir:
            file_handler = logging.FileHandler(os.path.join(logger_dir,
                                                            "training.log"))
            self.logger.addHandler(file_handler)

        console_handler = logging.StreamHandler()
        self.logger.addHandler(console_handler)

        self.iteration = 0

    def evaluate(self, batch):
        start = time.time()
        loss, accuracy, output = super().evaluate(batch)
        end = time.time()

        duration = end - start

        self.logger.info("Iteration: {}, Accuracy: {}, Loss: {}, Time: {}".format(self.iteration,
                                                                                  accuracy * 100,
                                                                                  loss,
                                                                                  duration))

        return loss, accuracy, output

    def run_epoch(self, sample_batch):
        """
        Runs an epoch and saves the parameters in a log

        :param sample_batch: A batch of input/label samples for training
        """
        start = time.time()
        loss = super().run_epoch(sample_batch)
        end = time.time()
        duration = end - start

        self.iteration += 1

        self.logger.info("Iteration: {}, Loss: {}, Time: {}".format(self.iteration,
                                                                    loss,
                                                                    duration))

        return loss

class ImageWriter(TrainerDecorator):
    """
    Write the image of each validation to a Tensorboard log
    """
    def __init__(self, trainer, logger_dir, experiment_name):
        """
        Initializes the Tensorboard writer

        :param trainer: Trainer object that the class wraps
        :param logger_dir: Directory to save Tensorboard logs
        :param experiment_name: The name to mark the experiment
        """
        if not os.path.isdir(logger_dir):
            raise IOError("{} is not a valid directory".format(logger_dir))

        super().__init__(trainer)
        experiment_dir = os.path.join(logger_dir, experiment_name)
        os.makedirs(experiment_dir, exist_ok=True)
        self.image_writer = tensorboardX.SummaryWriter(os.path.join(experiment_dir,
                                                       "validation_image"))

        self.iteration = 0

    def evaluate(self, batch):
        loss, accuracy, output = super().evaluate(batch)
        inputs = np.amax(batch[0].cpu().numpy(), axis=2).astype(np.float)
        inputs = (inputs + 200.0) * 0.50 / (np.max(inputs) - np.min(inputs))
        inputs = np.concatenate(list(inputs), axis=2)
        labels = np.amax(batch[1].cpu().numpy(), axis=2).astype(np.float) * 0.9
        labels = np.concatenate(list(labels), axis=2)
        prediction = np.amax(1/(1 + np.exp(-output[0])), axis=2)
        prediction = np.concatenate(list(prediction), axis=2)
        self.image_writer.add_image("input_image", inputs, self.iteration)
        self.image_writer.add_image("label_image", labels, self.iteration)
        self.image_writer.add_image("prediction_image", prediction, self.iteration)

        return loss, accuracy, output

    def run_epoch(self, sample_batch):
        """
        Runs an epoch and saves the parameters in a log

        :param sample_batch: A batch of input/label samples for training
        """
        loss = super().run_epoch(sample_batch)

        self.iteration += 1

        return loss



import itertools

import joblib
import numpy
import scipy.spatial
import skimage.morphology
import cc3d
import argschema

import cloudvolume
import kimimaro

from ac_segmentation.utils.tensorstore import open_tensor
from ac_segmentation.utils.io import (write_cv_skels_iter_tar)

np = numpy


def label_binary_array(binary_arr, size_threshold=200):
    
    labeled_arr, num_features = cc3d.connected_components(binary_arr, connectivity=6, return_N=True)
    if num_features > 1:
        labeled_arr = skimage.morphology.remove_small_objects(
            labeled_arr, min_size=size_threshold, connectivity=3, out=labeled_arr)
        
    return labeled_arr, len(np.unique(labeled_arr))


def threshold_binarize_array(arr, threshold=0.05):
    # dumb method
    return (arr >= threshold)


def skeletonize_labeled_array(
        labeled_arr, scale=2, constant=5,
        fill_holes=False, parallel=5,
        dust_threshold=10, **kwargs):

    skels = kimimaro.skeletonize(
        labeled_arr,
        teasar_params={
            "scale": scale, 
            "const": constant,  # influences the finger branches allowed
            "pdrf_scale": 10000,
            "pdrf_exponent": 1,
            "soma_acceptance_threshold": 3500,  # physical units
            "soma_detection_threshold": 750,  # physical units
            "soma_invalidation_const": 300,  # physical units
            "soma_invalidation_scale": 2,
            "max_paths": 50,  # default None
        },
        dust_threshold=dust_threshold,  # skip connected components with fewer than this many voxels
        anisotropy=(1, 1, 1),  # default True #influences the dimension scale
        fix_branching=True,  # default True
        fix_borders=True,  # default True
        fill_holes=fill_holes,  # default False
        fix_avocados=False,  # default False
        progress=False,  # default False, show progress bar
        parallel=parallel,  # <= 0 all cpu, 1 single process, 2+ multiprocess
        parallel_chunk_size=100,  # how many skeletons to process before updating progress bar
    )
    return skels


def join_components(skels, radius=2):
    # log which chunks the skeletons came from
    total_skels = []
    chunk_ind = []
    for ind, sk in enumerate(skels):
        temp_skels = sk.components()
        total_skels += temp_skels
        chunk_ind += len(temp_skels) * [ind]
        
    # extract root node coords
    root_vts = []  # node coordinates
    root_ind = []  # skeleton number from total_skels list
    root_node = []  # node associated with root_vt
    for ind, sk in enumerate(total_skels):  
        t_ids = sk.terminals()
        root_node += list(t_ids)
        ends = [sk.vertices[i] for i in t_ids]
        for end in ends:
            root_vts.append(list(end))
            root_ind.append(ind)
    
    # create kdtree and find all end nodes within given radius
    tree = scipy.spatial.KDTree(root_vts, leafsize=2)
    pairs = tree.query_pairs(radius)
    
    # get merge pair indices
    merge_pairs = []
    merge_pairs_vts = []
    for p1, p2 in pairs:
        merge_pairs.append([root_ind[p1], root_ind[p2]])
        merge_pairs_vts.append([root_vts[p1], root_vts[p2]])
    
    # check if skeletons in same chunk, if not merge
    fused_skels = []
    for ind, (m1, m2) in enumerate(merge_pairs):
        if chunk_ind[m1] == chunk_ind[m2]:
            continue
        else:
            try:
                fused = total_skels[m1].merge(total_skels[m2])
                v1, v2 = merge_pairs_vts[ind]
                n1, n2 = (fused.vertices.tolist().index(v1),
                          fused.vertices.tolist().index(v2))
                fused.edges = np.append(
                    fused.edges,
                    np.array([n1, n2]).reshape((1, 2)),
                    axis=0)
                total_skels[m1] = None
                total_skels[m2] = None
                fused_skels.append(fused)
            except Exception:
                pass
                
    out_skels = ([i for i in total_skels if i is not None] +
                 [i for i in fused_skels if i is not None])
    return out_skels


def skeletonize_labeled_array_concurrent(
        labeled_array, chunk_size=[1000, 1000, 1000],
        n_jobs=4, skel_search_radius=50):
    def skel_chunk(start, end):
        skels = skeletonize_labeled_array(
            np.array(labeled_array[
                start[0]:end[0],
                start[1]:end[1],
                start[2]:end[2]])
        )

        if len(skels) != 0:
            skels = [skel for skid, skel in skels.items()]
            skels = cloudvolume.Skeleton.simple_merge(skels).consolidate()
            skels.vertices += start
            return skels

    dx, dy, dz = labeled_array.shape
    xch, ych, zch = chunk_size
    sind_x, sind_y, sind_z = (
        list(range(0, dx, xch)),
        list(range(0, dy, ych)),
        list(range(0, dz, zch))
    )
    eind_x, eind_y, eind_z = (
        [x + xch for x in sind_x],
        [x + ych for x in sind_y],
        [x + zch for x in sind_z]
    )
    eind_x, eind_y, eind_z = (
        [dx if ele > dx else ele for ele in eind_x],
        [dy if ele > dy else ele for ele in eind_y],
        [dz if ele > dz else ele for ele in eind_z]
    )
    comb1 = list(itertools.product(sind_x, sind_y, sind_z))
    comb2 = list(itertools.product(eind_x, eind_y, eind_z))
    del sind_x, sind_y, sind_z, eind_x, eind_y, eind_z

    #skeletonize
    if len(comb1) == 1:
        res = skel_chunk(comb1[0], comb2[0])
    else:
        with joblib.parallel_config(backend="loky", inner_max_num_threads=1):
            res = joblib.Parallel(n_jobs=n_jobs)(joblib.delayed(skel_chunk)(x, y) for x, y in zip(comb1,comb2))

    #extract individual skeletons
    out_skels = []
    if res==None:
        pass
    elif isinstance(res,cloudvolume.skeleton.Skeleton):
        out_skels = res.components()
    else:
        res = [i for i in res if i is not None]
        for sk in res:
                out_skels += sk.components()

    return out_skels


def run(input_zarr, skeleton_output_path, probability_threshold=0.05,
        label_size_threshold=80, skeletonize_options=None):
    skeletonize_options = ({} if skeletonize_options is None
                           else skeletonize_options)
                           
    # Load segmentation
    prob_map = open_tensor(input_zarr).read().result()

    # binarize volume, label, and skeletonize
    binary_arr = threshold_binarize_array(
        prob_map, threshold=probability_threshold)
    labeled_arr, _ = label_binary_array(
        binary_arr, size_threshold=label_size_threshold)

    # skels = skeletonize_labeled_array(labeled_arr, **skeletonize_options)
    skels = skeletonize_labeled_array_concurrent(
        labeled_arr, **skeletonize_options
    )

    # write skeletons to swc zip
    # write_kimi_skels_tar(skeleton_output_path, skels)
    write_cv_skels_iter_tar(skeleton_output_path, skels)


class SkeletonizationOptions(argschema.schemas.DefaultSchema):
    n_jobs = argschema.fields.Int(required=False, allow_none=True)


class SkeletonizeZarrParameters(argschema.ArgSchema):
    input_zarr = argschema.fields.InputDir(required=True)
    skeleton_output = argschema.fields.OutputFile(required=True)
    probability_threshold = argschema.fields.Float(
        required=False, default=0.05)
    label_size_threshold = argschema.fields.Int(required=False, default=80)
    skeletonize_options = argschema.fields.Nested(
        SkeletonizationOptions, required=False, default=None, allow_none=True)


class SkeletonizeZarrModule(argschema.ArgSchemaParser):
    default_schema = SkeletonizeZarrParameters

    @property
    def skeletonize_options(self):
        return self.args["skeletonize_options"]    

    def run(self):
        run(self.args["input_zarr"],
            self.args["skeleton_output"],
            self.args["probability_threshold"],
            self.args["label_size_threshold"],
            self.skeletonize_options)


if __name__ == "__main__":
    mod = SkeletonizeZarrModule()
    mod.run()

__all__ = [
    "SkeletonizeZarrModule",
    "SkeletonizeZarrParameters"]



from ac_segmentation.neurotorch.datasets.dataset import AlignedVolume
from ac_segmentation.neurotorch.datasets.datatypes import Vector
from abc import abstractmethod
import numpy as np
from random import random


class Augmentation(AlignedVolume):
    def __init__(self, aligned_volume, iteration_size=None, stride=None,
                 frequency=1.0):
        self.setFrequency(frequency)
        self.setVolume(aligned_volume)
        self.setAugmentation(True)
        if iteration_size is None:
            iteration_size = self.getInputVolume().getIterationSize()
        if stride is None:
            stride = self.getInputVolume().getStride()
        self.setIteration(iteration_size, stride)

    def setAugmentation(self, augment):
        self.eval = augment

    def getAugmentation(self):
        return self.eval

    def get(self, bounding_box):
        if random() < self.frequency and self.getAugmentation():
            augmented_data = self.augment(bounding_box)
            return augmented_data
        else:
            data = (self.getInput(bounding_box), self.getLabel(bounding_box))
            return data

    def setFrequency(self, frequency=1.0):
        self.frequency = frequency

    def getBoundingBox(self):
        return self.getVolume().getBoundingBox()

    def setIteration(self, iteration_size, stride):
        self.getParent().setIterationSize(iteration_size)
        self.getParent().setStride(stride)

    def getInputVolume(self):
        return self.getVolume().getVolumes()[0]

    def getLabelVolume(self):
        return self.getVolume().getVolumes()[1]

    def getInput(self, bounding_box):
        return self.getInputVolume().get(bounding_box)

    def getLabel(self, bounding_box):
        return self.getLabelVolume().get(bounding_box)

    def setIterationSize(self, iteration_size):
        self.iteration_size = iteration_size

    def setStride(self, stride):
        self.stride = stride

    def setVolume(self, aligned_volume):
        self.aligned_volume = aligned_volume

    def getVolume(self):
        if isinstance(self.aligned_volume, AlignedVolume):
            return self.aligned_volume 
        if isinstance(self.aligned_volume, Augmentation) or \
           issubclass(type(self.aligned_volume), Augmentation):
            return self.aligned_volume.getVolume()

    def getVolumes(self):
        return self.getVolume().getVolumes()

    def getParent(self):
        return self.aligned_volume

    def __len__(self):
        return len(self.getParent())

    def __getitem__(self, idx):
        bounding_box = self.getParent()._indexToBoundingBox(idx)
        result = self.get(bounding_box)

        return result

    @abstractmethod
    def augment(self, bounding_box):
        pass

    def getValidData(self):
        return self.getParent().getValidData()



from ac_segmentation.utils.tensorstore import *

def zarr_to_CATMAID_project(zarr_path, out_path, container_id, AWS_Key, AWS_Secret_Key, project='NewProject', 
                            stack='NewStack', store='s3', chunks=(64,64,64), translation=(0,0,0), cutout=None, tile_dim=[128,128], ortho=False):

    #convert zarr to precomputed
    zarr_to_precomputed(zarr_path=zarr_path, AWS_Key=AWS_Key, AWS_Secret_Key=AWS_Secret_Key, 
                           out_path=out_path, store=store, chunks=(64,64,64), cutout=cutout)
    
    #extract url 
    split = out_path.split("/")
    bucket = split[0]
    key = out_path.replace(bucket+"/", '')
    url = os.path.join("https://", bucket+".s3-us-west-2.amazonaws.com", key, "%SCALE_DATASET%/")
    
    #get resolution and shape
    shape = open_tensor(zarr_path + "0/")[0,0,:,:,:].shape
    if cutout!=None:
        x1,x2,y1,y2,z1,z2 = cutout
        x,y,z = x2-x1,y2-y1,z2-z1
        shape = [x,y,z]
    r_path = os.path.join(os.path.dirname(os.path.dirname(zarr_path + "0/")), ".zattrs")
    res = json.loads(open(r_path, "r").read())['multiscales'][0]['datasets'][0]['coordinateTransformations'][0]['scale'][2:]
    
    #create project data json
    if ortho==True:
      stack_file = [{
        "project": {
          "title": project,
          "stacks": [{
            "title": stack+"xy",
            "dimension": str(tuple(shape)),
            "mirrors": [{
              "fileextension": "raw",
              "position": 0,
              "tile_source_type": 14,
              "tile_height":tile_dim[0],
              "tile_width":tile_dim[1],
              "title": stack+"_tiles",
              "url": url + "0_1_2"
            }],
            "resolution": str(tuple(res)),
            "translation": str(translation),
            "downsample_factors": ["(1,1,1)", "(2,2,2)", "(4,4,4)"],
            "orientation" : 0,
            "stackgroups": [{"title": project+"_StackGroup", "relation": "channel"}]
          },
          {
            "title": stack+"xz",
            "dimension": str(tuple([shape[0],shape[2],shape[1]])),
            "mirrors": [{
              "fileextension": "raw",
              "position": 0,
              "tile_source_type": 14,
              "tile_height":tile_dim[0],
              "tile_width":tile_dim[1],
              "title": stack+"_tiles",
              "url": url + "0_2_1"
            }],
            "resolution": str(tuple([res[0],res[2],res[1]])),
            "translation": str(tuple([translation[0],translation[2],translation[1]])),
            "downsample_factors": ["(1,1,1)", "(2,2,2)", "(4,4,4)"],
            "orientation" : 1,
            "stackgroups": [{"title": project+"_StackGroup", "relation": "view"}]
          },
          {
            "title": stack+"zy",
            "dimension": str(tuple([shape[2],shape[1],shape[0]])),
            "mirrors": [{
              "fileextension": "raw",
              "position": 0,
              "tile_source_type": 14,
              "tile_height":tile_dim[0],
              "tile_width":tile_dim[1],
              "title": stack+"_tiles",
              "url": url + "2_1_0"
            }],
            "resolution": str(tuple([res[2],res[1],res[0]])),
            "translation": str(tuple([translation[2],translation[1],translation[0]])),
            "downsample_factors": ["(1,1,1)", "(2,2,2)", "(4,4,4)"],
            "orientation" : 2,
            "stackgroups": [{"title": project+"_StackGroup", "relation": "view"}]
          }]
        }
      }]
      
    else:
      stack_file = [{
        "project": {
          "title": project,
          "stacks": [{
            "title": stack+"xy",
            "dimension": str(tuple(shape)),
            "mirrors": [{
              "fileextension": "raw",
              "position": 0,
              "tile_source_type": 14,
              "tile_height":tile_dim[0],
              "tile_width":tile_dim[1],
              "title": stack+"_tiles",
              "url": url + "0_1_2"
            }],
            "resolution": str(tuple(res)),
            "translation": str(translation),
            "downsample_factors": ["(1,1,1)", "(2,2,2)", "(4,4,4)"],
            "orientation" : 0,
            "stackgroups": [{"title": project+"_StackGroup", "relation": "channel"}]
          }]
        }
      }]
                                                              
    #save json to local                    
    json_fpath = os.path.join("./", "data.json")       
    with open(json_fpath, 'w') as f:
        json.dump(stack_file, f)
    
    #open docker container and import json
    client = docker.from_env()
    container = client.containers.get(container_id)
    copy_string = json_fpath + " " + container_id + ":" + "/home/django/projects/data.json"
    os.system("docker cp " + str(copy_string))
    container.exec_run('python3 manage.py catmaid_import_projects --input data.json --permission user:admin:can_import user:admin:can_annotate')

    #clean-up
    container.exec_run('rm input data.json')
    client.close()
    os.remove(json_fpath)

    return stack_file


import concurrent.futures
import gzip
import io
import tarfile
import numpy



def gzip_array(fn, arr):
    with gzip.open(fn, "wb") as f:
        numpy.save(f, arr)


def read_gzip_array(fn, preprocess_func=None):
    with gzip.open(fn, "rb") as f:
        x = numpy.load(f)
    if preprocess_func:
        x = preprocess_func(x)
    return x


# FIXME CL code does not preserve ids
def write_cv_skels_iter_tar(tar_fn, skels):
    with tarfile.open(tar_fn, mode="w:gz") as t:
        for skid, skel in enumerate(skels):
            bio = io.BytesIO(skel.to_swc().encode())
            info = tarfile.TarInfo(name=f"{skid}.swc")
            info.size = len(bio.getbuffer())
            t.addfile(tarinfo=info, fileobj=bio)



# Kisuk Lee <kisuklee@mit.edu>, 2017
# https://github.com/seung-lab/DeepEM
import numpy as np

def get_pair_first(arr, edge):
    shape = arr.size()[-3:]
    edge = np.array(edge)
    os1 = np.maximum(edge, 0)
    os2 = np.maximum(-edge, 0)
    ret = arr[..., os1[0]:shape[0]-os2[0],
                   os1[1]:shape[1]-os2[1],
                   os1[2]:shape[2]-os2[2]]
    return ret


def get_pair(arr, edge):
    shape = arr.size()[-3:]
    edge = np.array(edge)
    os1 = np.maximum(edge, 0)
    os2 = np.maximum(-edge, 0)
    arr1 = arr[..., os1[0]:shape[0]-os2[0],
                    os1[1]:shape[1]-os2[1],
                    os1[2]:shape[2]-os2[2]]
    arr2 = arr[..., os2[0]:shape[0]-os1[0],
                    os2[1]:shape[1]-os1[1],
                    os2[2]:shape[2]-os1[2]]
    return arr1, arr2


def crop_border(v, size):
    assert all([a > b for a, b in zip(v.shape[-3:], size[-3:])])
    sz, sy, sx = [s // 2 for s in size[-3:]]
    return v[..., sz:-sz, sy:-sy, sx:-sx]


def crop_center(v, size):
    assert all([a >= b for a, b in zip(v.shape[-3:], size[-3:])])
    z, y, x = size[-3:]
    sx = (v.shape[-1] - x) // 2
    sy = (v.shape[-2] - y) // 2
    sz = (v.shape[-3] - z) // 2
    return v[..., sz:sz+z, sy:sy+y, sx:sx+x]



import os
import numpy as np
import pandas as pd
import argschema as ags
import json
from operator import add
import itertools
from collections import deque, defaultdict
from scipy import ndimage as ndi
from skimage.morphology import remove_small_objects, skeletonize_3d
from sklearn.neighbors import KDTree
import tifffile as tif
from ac_segmentation.deprecated.swc_morphology.classes import Morphology
from ac_segmentation.deprecated.swc_morphology.readwrite_swc import morphology_from_swc, morphology_to_swc

class AddRadiusParameters(ags.ArgSchema):
    input_file = ags.fields.InputFile(required=True, description='Input file')
    output_file = ags.fields.OutputFile(required=True, description='Output file')
    csv_file = ags.fields.InputFile(required=True, description='Segmentation file')
    pxl_xy = ags.fields.Int(dtype='int', required=False, default=812, 
                                desctiption='pxl xy size in nm') 

def process(infile, outfile, csvfile, pxl_xy):
    df = pd.read_csv(csvfile)
    arr = df.values
    morph_in = morphology_from_swc(infile)
    dict_out = add_radius_to_morph_10x(morph_in, arr, pxl_xy)
    morphology_to_swc(dict_out['morph'], outfile)              

def add_radius_to_morph_10x(morph, non_specific_segmentation, pxl_xy, 
                            wiggle_room = 50, **kwargs):
    """
    This function will iterate through each node in swc file and get 10 nodes up and 10 nodes down.
    With this segment a bounding box is created. Wiggle room is added to x-y dimensions because the 
    20 node segment is only a skeleton structure wheras the segmentation.csv will be much wider. 
    Wiggle_room is the padding for this bounding box, 80 pixels left and 80 pixels right of the 
    max and min x/y values. Only one z-slice is taken as we are only considering x-y for radius calculations

    """
    # Modified version to calculate radius only for every 10th node
    mod_morph = morph.clone()
    numnodes = len(morph.nodes())
    print("     {} Nodes to add. Estimated time to complete = {} minutes".format(numnodes,numnodes/1000))
    empty_ct = 0
    node_ct = 0
    for node in [n for n in mod_morph.nodes() if n['type'] != 1]:
        if node_ct % 10 == 0:
            node_coord = (node['x'],node['y'])
            coords_up_and_down = n_nodes_up_and_down(node,10,mod_morph)
            xyz_coords = np.asarray(list(coords_up_and_down))

            min_bb = [int(min(xyz_coords[:,j])) for j in range(0,3)]
            max_bb = [int(max(xyz_coords[:,j])) for j in range(0,3)] 
            inside_bbox = bounding_box(non_specific_segmentation, 
                          min_x = min_bb[0] - wiggle_room , max_x = max_bb[0] + wiggle_room,
                          min_y = min_bb[1] - wiggle_room , max_y = max_bb[1] + wiggle_room,
                          min_z = min_bb[2] - 1 , max_z = max_bb[2] + 1)
            segmented_local_xyz_array = non_specific_segmentation[inside_bbox]
            segmented_local_xy_array = segmented_local_xyz_array[:,0:2]

            if segmented_local_xy_array.size != 0:
                local_segmentation_lookup_tree_raw = KDTree(segmented_local_xy_array,leaf_size=40)
                dist_stepper=0
                condition=0
                explored = []
                while condition != 1:
                    dist_stepper+=1
                    movement_vectors = [p for p in itertools.product([n for n in range(1,dist_stepper+1)]+[-n for n in range(0,dist_stepper+1)], repeat=2) if p != (0,0)]
                    x = {v:((v[0]**2)+(v[1]**2))**0.5 for v in movement_vectors}
                    ordered_dict = {k: v for k, v in sorted(x.items(), key=lambda item: item[1])}
                    offsets=[]
                    for offset in ordered_dict.keys():
                        node_to_check = np.array([sum(x) for x in zip(offset,node_coord)])
                        explored.append(node_to_check)
                        dist, _ = local_segmentation_lookup_tree_raw.query(node_to_check.reshape(1,2), k=1)
                        dist = dist[0][0]
                        offsets.append(offset)
                        if dist !=0:
                            distance = ((offset[0]**2)+(offset[1]**2))**0.5
                            condition = 1
                            break
                node['radius'] = distance*pxl_xy/1000 #nm->um

            else:
                empty_ct+=1
                node['radius'] = 0.1
            r = node['radius']    
        else:
            node['radius'] = r
        node_ct += 1
        
    add_missing_radius_vals(mod_morph)
    
    result_dict = {}
    result_dict['morph'] = mod_morph
    return result_dict

def n_nodes_up_and_down(st_node,n,morph):
    return n_nodes_up(st_node,n,morph).union(n_nodes_down(st_node,n,morph))

def n_nodes_up(st_node,n,morph):
    
    cur_node = st_node
    ct=0
    nodes_up = set()
    nodes_up.add((int(st_node['x']),int(st_node['y']),int(st_node['z'])))
    while ct != n:
        
        parent_id = cur_node['parent']
        if parent_id == -1:
            return nodes_up
        else:
            next_node = morph.node_by_id(parent_id)
            nodes_up.add((int(next_node['x']),int(next_node['y']),int(next_node['z'])))
            cur_node = next_node
        ct+=1
    return nodes_up

def n_nodes_down(node,n,morph):
    ct = 0
    nodes_down = set()
    while n != ct:
        
        next_node = morph.get_children(node)
        if next_node !=[]:
            for node in next_node:
                nodes_down.add((int(node['x']),int(node['y']),int(node['z'])))
        else:
            return nodes_down
        ct+=1

    return nodes_down

def bounding_box(points, min_x=-np.inf, max_x=np.inf, min_y=-np.inf,
                        max_y=np.inf, min_z=-np.inf, max_z=np.inf):
    bound_x = np.logical_and(points[:, 0] > min_x, points[:, 0] < max_x)
    bound_y = np.logical_and(points[:, 1] > min_y, points[:, 1] < max_y)
    bound_z = np.logical_and(points[:, 2] > min_z, points[:, 2] < max_z)

    bb_filter = np.logical_and(np.logical_and(bound_x, bound_y), bound_z)

    return bb_filter

def add_missing_radius_vals(radius_morph):
    """
    Because some nodes were added during postprocessing (i.e connection algorithm)
    They may not be in the segmentation.csv This script finds the nearest node up
    or down stream that has a radius value calculated. 
    Defaults to tree averages if the above fails
    """
    for missing_rad_node in [n for n in radius_morph.nodes() if n['radius'] == 0.1]:
        try:
            curr_node = missing_rad_node
            up_down_dict = defaultdict(dict)
            upstep =0
            while curr_node['radius'] == 0.1:
                upstep+=1
                curr_node = radius_morph.node_by_id(curr_node['parent'])
            up_down_dict['up']['steps'] = upstep
            up_down_dict['up']['radius'] = curr_node['radius']


            queue = deque([missing_rad_node['id']])
            curr_node_down_id = missing_rad_node['id']
            downstep=0
            while radius_morph.node_by_id(curr_node_down_id)['radius'] == 0.1:
                downstep+=1
                curr_node_down_id = queue.popleft()
                for ch_no in radius_morph.get_children(radius_morph.node_by_id(curr_node_down_id)):
                    queue.append(ch_no['id'])

            up_down_dict['down']['steps'] = downstep
            up_down_dict['down']['radius'] = radius_morph.node_by_id(curr_node_down_id)['radius']
            closest_direction = [k for k,v in up_down_dict.items() if v['steps'] == min([s['steps'] for s in up_down_dict.values() ])][0]

            missing_rad_node['radius'] = up_down_dict[closest_direction]['radius']
        except:
            missing_rad_node['radius'] = np.mean([n['radius'] for n in radius_morph.nodes() if n['type']!=1])
    
def add_radius_to_morph(morph, non_specific_segmentation, pxl_xy,
                        wiggle_room = 50,**kwargs):
    """
    This function will iterate through each node in swc file and get 10 nodes up and 10 nodes down.
    With this segment a bounding box is created. Wiggle room is added to x-y dimensions because the 
    20 node segment is only a skeleton structure wheras the segmentation.csv will be much wider. 
    Wiggle_room is the padding for this bounding box, 80 pixels left and 80 pixels right of the 
    max and min x/y values. Only one z-slice is taken as we are only considering x-y for radius calculations

    """
    mod_morph = morph.clone()
    numnodes = len(morph.nodes())
    print("     {} Nodes to add. Estimated time to complete = {} minutes".format(numnodes,numnodes/1000))
    empty_ct = 0
    for node in [n for n in mod_morph.nodes() if n['type'] != 1]:
        node_coord = (node['x'],node['y'])
        coords_up_and_down = n_nodes_up_and_down(node,10,mod_morph)
        xyz_coords = np.asarray(list(coords_up_and_down))
        
        min_bb = [int(min(xyz_coords[:,j])) for j in range(0,3)]
        max_bb = [int(max(xyz_coords[:,j])) for j in range(0,3)] 
        inside_bbox = bounding_box(non_specific_segmentation, 
                      min_x = min_bb[0] - wiggle_room , max_x = max_bb[0] + wiggle_room,
                      min_y = min_bb[1] - wiggle_room , max_y = max_bb[1] + wiggle_room,
                      min_z = min_bb[2] - 1 , max_z = max_bb[2] + 1)
        segmented_local_xyz_array = non_specific_segmentation[inside_bbox]
        segmented_local_xy_array = segmented_local_xyz_array[:,0:2]

        if segmented_local_xy_array.size != 0:
            local_segmentation_lookup_tree_raw = KDTree(segmented_local_xy_array,leaf_size=40)
            dist_stepper=0
            condition=0
            explored = []
            while condition != 1:
                dist_stepper+=1
                movement_vectors = [p for p in itertools.product([n for n in range(1,dist_stepper+1)]+[-n for n in range(0,dist_stepper+1)], repeat=2) if p != (0,0)]
                x = {v:((v[0]**2)+(v[1]**2))**0.5 for v in movement_vectors}
                ordered_dict = {k: v for k, v in sorted(x.items(), key=lambda item: item[1])}
                offsets=[]
                for offset in ordered_dict.keys():
                    node_to_check = np.array([sum(x) for x in zip(offset,node_coord)])
                    explored.append(node_to_check)
                    dist, _ = local_segmentation_lookup_tree_raw.query(node_to_check.reshape(1,2), k=1)
                    dist = dist[0][0]
                    offsets.append(offset)
                    if dist !=0:
                        distance = ((offset[0]**2)+(offset[1]**2))**0.5
                        condition = 1
                        break
            node['radius'] = distance*pxl_xy/1000 #nm->um

        else:
            empty_ct+=1
            node['radius'] = 0.1
    add_missing_radius_vals(mod_morph)
    
    result_dict = {}
    result_dict['morph'] = mod_morph
    return result_dict
    
def n_nodes_up_and_down(st_node,n,morph):
    return n_nodes_up(st_node,n,morph).union(n_nodes_down(st_node,n,morph))

def n_nodes_up(st_node,n,morph):
    
    cur_node = st_node
    ct=0
    nodes_up = set()
    nodes_up.add((int(st_node['x']),int(st_node['y']),int(st_node['z'])))
    while ct != n:
        
        parent_id = cur_node['parent']
        if parent_id == -1:
            return nodes_up
        else:
            next_node = morph.node_by_id(parent_id)
            nodes_up.add((int(next_node['x']),int(next_node['y']),int(next_node['z'])))
            cur_node = next_node
        ct+=1
    return nodes_up

def n_nodes_down(node,n,morph):
    ct = 0
    nodes_down = set()
    while n != ct:
        
        next_node = morph.get_children(node)
        if next_node !=[]:
            for node in next_node:
                nodes_down.add((int(node['x']),int(node['y']),int(node['z'])))
        else:
            return nodes_down
        ct+=1

    return nodes_down

def bounding_box(points, min_x=-np.inf, max_x=np.inf, min_y=-np.inf,
                        max_y=np.inf, min_z=-np.inf, max_z=np.inf):
    bound_x = np.logical_and(points[:, 0] > min_x, points[:, 0] < max_x)
    bound_y = np.logical_and(points[:, 1] > min_y, points[:, 1] < max_y)
    bound_z = np.logical_and(points[:, 2] > min_z, points[:, 2] < max_z)

    bb_filter = np.logical_and(np.logical_and(bound_x, bound_y), bound_z)

    return bb_filter
    
class AddRadius(ags.ArgSchemaParser):
        
    def run(self):
        process(self.args['input_file'], self.args['output_file'], self.args['csv_file'],
                self.args['pxl_xy'])
   
        
if __name__ == "__main__":
    mod = AddRadius(schema_type=AddRadiusParameters)
    mod.run()      


import os
import numpy as np
import pandas as pd
import argschema as ags
import json
import tifffile as tif

import ac_segmentation.neurotorch.datasets.datatypes
import ac_segmentation.neurotorch.datasets.dataset
import ac_segmentation.neurotorch.nets.RSUNet
import ac_segmentation.neurotorch.core.predictor
import ac_segmentation.neurotorch.datasets.filetypes

Predictor = ac_segmentation.neurotorch.core.predictor.Predictor
Vector = ac_segmentation.neurotorch.datasets.datatypes.Vector
BoundingBox = ac_segmentation.neurotorch.datasets.datatypes.BoundingBox
Array = ac_segmentation.neurotorch.datasets.dataset.Array
RSUNet = ac_segmentation.neurotorch.nets.RSUNet.RSUNet
TiffVolume = ac_segmentation.neurotorch.datasets.filetypes.TiffVolume



class SegmentationParameters(ags.ArgSchema):
    ckpt = ags.fields.InputFile(required=True, description='Model file')
    output_dir = ags.fields.OutputDir(required=True, description='Output directory')
    chunk_size = ags.fields.Int(dtype='int', required=False, default=416)
    overlap = ags.fields.Int(dtype='int', required=False, default=16)

def predict(checkpoint, outdir, mip=0, **kwargs):
    
    #set default keyword arguments
    defaultKwargs = {'iter_size': BoundingBox(Vector(0, 0, 0), Vector(128, 128, 32)) , 'stride': Vector(64, 64, 16),
                    'gpu_device': 0, 'batch_size':20}
    kwargs = { **defaultKwargs, **kwargs }
    
    #set number of chunks and load/set bounding box
    num_chunks = len([f for f in os.listdir(os.path.join(outdir, 'inputs')) 
                        if f.startswith('chunk')])
    df = pd.read_csv(os.path.join(outdir, 'inputs','bbox.csv'))
    bb = df.bounding_box.values
    bbn = BoundingBox(Vector(bb[0], bb[1], bb[2]), Vector(bb[3], bb[4], bb[5]))
        
    # Initialize the U-Net architecture
    net = RSUNet()
       
    for n in range(num_chunks):
        filename = os.path.join(outdir, 'inputs', 'chunk%02d.tif'%n)
        print('chunk%02d.tif'%n) 
        
        with TiffVolume(filename, bbn, kwargs['iter_size'], kwargs['stride']) as inputs:              
            # Predict
            predictor = Predictor(net, checkpoint, gpu_device=kwargs['gpu_device']) 
            output_volume = Array(-np.inf*np.ones(inputs.getBoundingBox().getNumpyDim(), dtype=np.float32))
            predictor.run(inputs, output_volume, batch_size=kwargs['batch_size'])
                            
            convert_prob_map(outdir, output_volume, n)
            
    return num_chunks
    
def predict_array(checkpoint, outdir, inarr, mip=0, **kwargs):
    
    #set default keyword arguments
    defaultKwargs = {'output_type': 'volume', 'iter_size': BoundingBox(Vector(0, 0, 0), Vector(128, 128, 32)), 'stride': Vector(64, 64, 16),
                    'gpu_device': None, 'batch_size':20}
    kwargs = { **defaultKwargs, **kwargs }
    
    #set array object
    inarr = Array(inarr, iteration_size= kwargs['iter_size'], stride= kwargs['stride'])
        
    # Initialize the U-Net architecture
    net = RSUNet()
    
    # Predict
    predictor = Predictor(net, checkpoint, gpu_device=kwargs['gpu_device']) 
    output_volume = Array(-np.inf*np.ones(inarr.getBoundingBox().getNumpyDim(), dtype=np.float32))
    predictor.run(inarr, output_volume, batch_size=kwargs['batch_size'])
    
    if kwargs['output_type'] == 'volume':
        return output_volume
    
    else:
        convert_prob_map(outdir, output_volume, 1)
        

def convert_prob_map(outdir, output_volume, chunk_n):
    savedir = os.path.join(outdir, 'Segmentation')
    if not os.path.isdir(savedir):
        os.mkdir(savedir)
    mips_dir = os.path.join(outdir, 'Segmentation', 'mips')   
    if not os.path.isdir(mips_dir):
        os.mkdir(mips_dir)
            
    # Convert to probability map and save
    probability_map = 1/(1+np.exp(-output_volume.getArray()))
    chunkdir = os.path.join(savedir, 'chunk%02d'%chunk_n)

    if not os.path.isdir(chunkdir):
        os.mkdir(chunkdir)
    for i in range(probability_map.shape[0]):# save 8-bit version as multiple tif files
        tif.imsave(os.path.join(chunkdir,'%03d.tif'%i), np.uint8(255*probability_map[i,:,:]))          
    mip_xy = np.max(probability_map,0) # save mip
    tif.imsave(os.path.join(mips_dir, 'MAX_xy_chunk%02d.tif'%chunk_n), np.uint8(255*mip_xy)) 
    
def load_stack_select(dirname, idx1, idx2):
    # Load image stack filenames
    filelist = [f for f in os.listdir(dirname) if f.endswith('.tif')] 
    filelist.sort()
    
    # Calculate stack size
    filename = os.path.join(dirname, filelist[0])
    img = tif.imread(filename)
    cell_stack_size = idx2 - idx1 + 1, img.shape[0], img.shape[1]    
    stack = np.zeros(cell_stack_size, dtype=img.dtype)
    for i, idx in enumerate(range(idx1, idx2 + 1)):
        filename = os.path.join(dirname, filelist[idx])
        img = tif.imread(filename)
        stack[i,:,:] = img
        
    return stack                
            
def blend_chunks(outdir, num_chunks, chunk_size, overlap):
    savedir = os.path.join(outdir, 'Segmentation')
    for n1 in range(num_chunks - 1):
        n2 = n1 + 1
        # Load overlapping tif files
        indir1 = os.path.join(savedir, 'chunk%02d'%n1)
        indir2 = os.path.join(savedir, 'chunk%02d'%n2)
        idx11 = chunk_size - overlap
        idx12 = chunk_size - 1
        stack1 = load_stack_select(indir1, idx11, idx12)
        idx21 = 0
        idx22 = overlap - 1
        stack2 = load_stack_select(indir2, idx21, idx22)
        # Blend overlapping images
        shift_x = (chunk_size - overlap)*4
        stack2_shift = np.zeros((stack1.shape), dtype=stack1.dtype)
        stack2_shift[:,:,0:stack1.shape[2] - shift_x] = stack2[:,:,shift_x:]
        stack_blended = np.maximum(stack1, stack2_shift)
        # Save blended images
        for i, idx in enumerate(range(idx11, idx12 + 1)):
            tif.imsave(os.path.join(indir1, '%03d.tif'%idx), stack_blended[i,:,:])
        stack_blended_shift = np.zeros((stack1.shape), dtype=stack1.dtype)
        stack_blended_shift[:,:,shift_x:] = stack_blended[:,:,0:stack1.shape[2] - shift_x] 
        for i, idx in enumerate(range(idx21, idx22 + 1)):
            tif.imsave(os.path.join(indir2, '%03d.tif'%idx), stack_blended_shift[i,:,:])
            
class Segmentation(ags.ArgSchemaParser):
        
    def run(self):
        num_chunks = predict(self.args['ckpt'], self.args['output_dir'],
            chunk_size=self.args['chunk_size'], overlap = self.args['overlap']) 
        blend_chunks(self.args['output_dir'], num_chunks, 
            chunk_size=self.args['chunk_size'], overlap = self.args['overlap'])     
        
if __name__ == "__main__":
    mod = Segmentation(schema_type=SegmentationParameters)
    mod.run()      


import glob
import os
import pytest

def get_submod(directory, filetype):
    submods = []
    for file in glob.glob(f"{directory}/**/{filetype}", recursive=True):
        submods.append(file)
    return submods                 

#get submodules and reformat
path = os.path.normpath(os.path.dirname(os.path.realpath(__file__)) + os.sep + os.pardir)+"/src/ac_segmentation/"
matches = ['init','neurotorch','test']
submods = get_submod(path, '*.py')
submods = [x.split('src/')[1].replace('/','.').replace('.py','') for x in submods]
submods = [ x for x in submods if any(y in x for y in matches)==False]

@pytest.mark.parametrize("mod", submods)
def test_import(mod):
    imp = __import__(mod)
      


import pytest
import numpy as np
import ac_segmentation
from ac_segmentation.utils.tensorstore import create_tensor
from ac_segmentation.utils.io import gzip_array


@pytest.fixture(scope="session")
def zarr_file(tmp_path_factory):
    fn = str(tmp_path_factory.mktemp("temp"))
    arr = np.random.choice([0, 1], size=(1,1,100,100,100), p=[.7, .3])
    tensor = create_tensor(fpath=fn+"test_zarr.zarr", arr_shape=[1,1,100,100,100], arr=arr, chunk_shape = [1,1,64,64,64], dtype = 'uint16', fill_value=0)
    return fn
    
    
@pytest.fixture(scope="session")
def npy_file(tmp_path_factory):
    fn = str(tmp_path_factory.mktemp("temp"))
    arr = np.random.choice([0, 1], size=(100,100,100), p=[.7, .3])
    gzip_array(fn+"test_zarr.npy.gz", arr)
    return fn


#!/usr/bin/env python
__doc__ = """

Symmetric 3d U-Net implemented in PyTorch

(Optional)
Factorized 3D convolution
Extra residual connections

Nicholas Turner <nturner@cs.princeton.edu>, 2017
Based on an architecture by
Kisuk Lee <kisuklee@mit.edu>, 2017
"""

from torch import nn
from torch.nn import functional as F
import ac_segmentation.neurotorch.nets.layers as layers
from ac_segmentation.neurotorch.nets.netcollector import NetCollector
from collections import OrderedDict


# Global switches
factorize = False
residual = True
bn = True

# Number of feature maps
nfeatures = [16, 16, 64, 128, 80, 96]

# Filter size
sizes = [(3, 3, 3),
         (3, 3, 3),
         (3, 3, 3),
         (3, 3, 3),
         (3, 3, 3),
         (3, 3, 3)]

# In/out filter & stride size
io_size = (3, 3, 3)
io_stride = (1, 1, 1)


class ConvMod(nn.Module):
    """ Convolution "module" """

    def __init__(self, D_in, D_out, ks, activation=F.elu,
                 fact=factorize, resid=residual,
                 bn=bn, momentum=0.5):

        nn.Module.__init__(self)
        st = (1, 1, 1)
        pd = layers.pad_size(ks, "same")
        # conv layer constructor
        conv_constr = layers.FactConv if fact else layers.Conv
        bias = not bn

        self.resid = resid
        self.bn = bn
        self.activation = activation

        first_pd = layers.pad_size((1, ks[1], ks[2]), "same")
        self.conv1 = conv_constr(
            D_in, D_out, (1, ks[1], ks[2]), st, first_pd, bias)
        self.conv2 = conv_constr(D_out, D_out, ks, st, pd, bias)
        self.conv3 = conv_constr(D_out, D_out, ks, st, pd, bias)

        if self.bn:
            self.bn1 = nn.BatchNorm3d(D_out, momentum=momentum)
            self.bn2 = nn.BatchNorm3d(D_out, momentum=momentum)
            self.bn3 = nn.BatchNorm3d(D_out, momentum=momentum)

    def forward(self, x):

        out1 = self.conv1(x)
        if self.bn:
            out1 = self.bn1(out1)
        out1 = self.activation(out1)

        out2 = self.conv2(out1)
        if self.bn:
            out2 = self.bn2(out2)
        out2 = self.activation(out2)

        out3 = self.conv3(out2)

        if self.resid:
            out3 = out3 + out1

        if self.bn:
            return self.activation(self.bn3(out3))
        else:
            return self.activation(out3)


class ConvTMod(nn.Module):
    """ Transposed Convolution "module" """

    def __init__(self, D_in, D_out, ks, up=(2, 2, 2), activation=F.elu,
                 fact=factorize, resid=residual,
                 bn=bn, momentum=0.5):

        nn.Module.__init__(self)

        # ConvT constructor
        convt_constr = layers.FactConvT if fact else layers.ConvT
        self.bn = bn
        self.activation = activation
        bias = not bn

        self.convt = convt_constr(D_in, D_out, ks=up, st=up, bias=bias)
        if bn:
            self.bn1 = nn.BatchNorm3d(D_out, momentum=momentum)

        self.convmod = ConvMod(D_out, D_out, ks, fact=fact, resid=resid, bn=bn)

    def forward(self, x, skip):

        if self.bn:
            convt = self.activation(self.bn1(self.convt(x) + skip))
        else:
            convt = self.activation(self.convt(x) + skip)

        return self.convmod(convt)


class Conv(nn.Module):
    """ Single convolution module """

    def __init__(self, D_in, D_out, ks, st=(1, 1, 1), activation=F.elu,
                 fact=factorize):

        nn.Module.__init__(self)
        pd = layers.pad_size(ks, "same")

        conv_constr = layers.FactConv if fact else layers.Conv
        self.activation = activation

        self.conv = conv_constr(D_in, D_out, ks, st, pd, bias=True)

    def forward(self, x):
        return self.activation(self.conv(x))


class OutputModule(nn.Module):
    """ Hidden representation -> Output module """

    def __init__(self, D_in, outspec, ks=io_size, st=io_stride):
        """ outspec should be an Ordered Dict """

        nn.Module.__init__(self)

        pd = layers.pad_size(ks, "same")

        self.output_layers = []
        for (name, d_out) in outspec.items():
            setattr(self, name, layers.Conv(
                D_in, d_out, ks, st, pd, bias=True))
            self.output_layers.append(name)

    def forward(self, x):
        return [getattr(self, layer)(x) for layer in self.output_layers]


class RSUNet(nn.Module):
    """ Full model """

    def __init__(self, D_in=1, output_spec=OrderedDict(soma_label=1),
                 depth=4, io_size=io_size,
                 io_stride=io_stride, bn=bn):

        nn.Module.__init__(self)

        assert depth < len(nfeatures)
        self.depth = depth

        # D_in represents the input dimension (#feature maps)
        # in most pytorch docs. I'll follow that convention here

        # Input feature embedding without batchnorm
        fs = nfeatures[0]
        self.inputconv = Conv(D_in, fs, io_size, st=io_stride)
        D_in = fs

        # modules within up/down pathways
        # are added with setattr calls
        # slightly obscured by U3D methods

        # Contracting pathway
        for d in range(depth):
            fs = nfeatures[d]
            ks = sizes[d]
            self.add_conv_mod(d, D_in, fs, ks, bn)
            self.add_max_pool(d+1, fs)
            D_in = fs

        # Bridge
        fs = nfeatures[depth]
        ks = sizes[depth]
        self.add_conv_mod(depth, D_in, fs, ks, bn)
        D_in = fs

        # Expanding pathway
        for d in reversed(range(depth)):
            fs = nfeatures[d]
            ks = sizes[d]
            self.add_deconv_mod(d, D_in, fs, bn, ks)
            D_in = fs

        # Output feature embedding without batchnorm
        self.embedconv = Conv(D_in, D_in, ks, st=(1, 1, 1))

        # Output by spec
        self.outputdeconv = OutputModule(
            D_in, output_spec, ks=io_size, st=io_stride)

    def add_conv_mod(self, depth, D_in, D_out, ks, bn):

        setattr(self, "convmod{}".format(depth),
                ConvMod(D_in, D_out, ks, bn=bn))

    def add_max_pool(self, depth, D_in, down=(2, 2, 2)):

        setattr(self, "maxpool{}".format(depth),
                nn.MaxPool3d(down))

    def add_deconv_mod(self, depth, D_in, D_out, bn, up=(2, 2, 2)):

        setattr(self, "deconv{}".format(depth),
                ConvTMod(D_in, D_out, up, bn=bn))

    def forward(self, x):

        # Input feature embedding without batchnorm
        x = self.inputconv(x)

        # Contracting pathway
        skip = []
        for d in range(self.depth):
            convmod = getattr(self, "convmod{}".format(d))
            maxpool = getattr(self, "maxpool{}".format(d+1))
            cd = convmod(x)
            x = maxpool(cd)

            skip.append(cd)

        # Bridge
        bridge = getattr(self, "convmod{}".format(self.depth))
        x = bridge(x)

        # Expanding pathway
        for d in reversed(range(self.depth)):
            deconv = getattr(self, "deconv{}".format(d))
            x = deconv(x, skip[d])

        # Output feature embedding without batchnorm
        x = self.embedconv(x)

        return self.outputdeconv(x)


NetCollector().add_module(RSUNet(), "RSUNet")



import numpy as np

def lut_preprocess_array(arr, max_int):
    lut = np.empty(arr.max() + max_int, dtype="uint8")
    lut[max_int:] = 255
    lut[:max_int] = np.round(np.arange(max_int) * (255 / max_int))
    return lut[arr]


import numpy
import torch
import argschema

try:
    import zarr
except ImportError:
    pass
    
from ac_segmentation.utils.io import (
    gzip_array, write_cv_skels_iter_tar
)

import ac_segmentation
import ac_segmentation.neurotorch.datasets.datatypes
import ac_segmentation.neurotorch.datasets.dataset
import ac_segmentation.neurotorch.nets.RSUNet
import ac_segmentation.neurotorch.core.predictor
from ac_segmentation.utils.tensorstore import open_ZarrTensor
from ac_segmentation.utils.preprocess import lut_preprocess_array


Predictor = ac_segmentation.neurotorch.core.predictor.Predictor
Vector = ac_segmentation.neurotorch.datasets.datatypes.Vector
BoundingBox = ac_segmentation.neurotorch.datasets.datatypes.BoundingBox
TSArray = ac_segmentation.neurotorch.datasets.dataset.TSArray
Array = ac_segmentation.neurotorch.datasets.dataset.Array
RSUNet = ac_segmentation.neurotorch.nets.RSUNet.RSUNet
np = numpy

ONE_GiB = 1_000_000_000


# hack to set number of threads w/ OMP environment variable for old torch
import os
if os.getenv("OMP_NUM_THREADS"):
    torch.set_num_threads(int(os.getenv("OMP_NUM_THREADS")))


def predict_array(
        weights_file, arr,
        iter_size=BoundingBox(Vector(0, 0, 0), Vector(64, 64, 64)),
        stride=Vector(32, 32, 32),
        batch_size=80, gpu_device=None):
    inarr = Array(arr, iteration_size=iter_size, stride=stride)
    outarr = Array(
        -np.inf * np.ones(
            inarr.getBoundingBox().getNumpyDim(), dtype=np.float32),
        iteration_size=iter_size, stride=stride)
    net = ac_segmentation.neurotorch.nets.RSUNet.RSUNet()
    predictor = Predictor(net, weights_file, gpu_device=gpu_device)
    predictor.run(inarr, outarr, batch_size=batch_size)

    # prob_map = 1/(1+np.exp(-outarr.getArray()))
    prob_map = torch.special.expit(
        torch.from_numpy(outarr.getArray())
    ).numpy()
    return prob_map


# TODO predict_arr_chunked function


def predict_zarr(zarr_loc, weights_file, level=0,
                 max_intensity=30000, **kwargs):
    z = zarr.load(zarr_loc)
    ds = z[level]
    data = numpy.transpose(ds[0, 0, ...])
    data = lut_preprocess_array(data, max_intensity)

    prob_arr = predict_array(weights_file, data, **kwargs)
    return numpy.transpose(prob_arr)


def predict_zarr_ts(zarr_loc, weights_file, level=0,
                    max_intensity=30000, bytes_limit=(5 * ONE_GiB),
                    iter_size=BoundingBox(Vector(0, 0, 0), Vector(64, 64, 64)),
                    stride=Vector(32, 32, 32),
                    batch_size=80, gpu_device=None, bound_box=None):
    try:
      in_ts = open_ZarrTensor(f"{zarr_loc}/{level}", bytes_limit=bytes_limit)
    except:
      in_ts = open_ZarrTensor(zarr_loc, bytes_limit=bytes_limit)
      
    in_ts = in_ts[0,0,...].transpose()
    
    if bound_box:
        x1,x2,y1,y2,z1,z2 = bound_box
        in_ts = in_ts[z1:z2,y1:y2,x1:x2]
    
    in_arr = TSArray(in_ts, iteration_size=iter_size, stride=stride)
    out_arr = Array(
        -np.inf * np.ones(in_ts.shape,
                          dtype=np.float32),
        iteration_size=iter_size, stride=stride)
    net = ac_segmentation.neurotorch.nets.RSUNet.RSUNet()
    predictor = Predictor(net, weights_file, gpu_device=gpu_device)
    predictor.run(
        in_arr, out_arr, batch_size=batch_size, max_pix=max_intensity)

    # prob_map = 1/(1+np.exp(-outarr.getArray()))
    prob_map = torch.special.expit(
        torch.from_numpy(
            out_arr.getArray())
    ).numpy()
    return prob_map.transpose()
    
    
        
def run(weights_file, input_zarr, probability_output_path,
        zarr_level=0, filter_max_intensity=30000,
        predict_options=None, skeletonize_options=None):
        
    # predict and return as probability
    prob_map = predict_zarr_ts(
        input_zarr, weights_file, level=zarr_level,
        max_intensity=filter_max_intensity, gpu_device=predict_options['gpu_device'], 
        bound_box=predict_options['bound_box'], batch_size=predict_options['batch_size'])

    # write out uint8 representation of probabilities
    uint8_prob_map = (prob_map * 255).astype(numpy.uint8)
    gzip_array(probability_output_path, uint8_prob_map)
    del uint8_prob_map


class PredictOptions(argschema.schemas.DefaultSchema):
    gpu_device = argschema.fields.Int(required=False, allow_none=True, default=None)
    batch_size = argschema.fields.Int(required=False, allow_none=True, default=80)
    bound_box = argschema.fields.List(argschema.fields.Int(),required=False, default='', allow_none=True)


class SegmentZarrParameters(argschema.ArgSchema):
    input_zarr = argschema.fields.InputDir(required=True)
    weights_file = argschema.fields.InputFile(required=True)
    probability_output = argschema.fields.OutputFile(required=True)

    zarr_level = argschema.fields.Int(required=False, default=1)
    filter_max_intensity = argschema.fields.Int(required=False, default=30000)
    predict_options = argschema.fields.Nested(
        PredictOptions, required=False,
        default=None, allow_none=True)

    output_json = argschema.fields.OutputFile(required=False, allow_none=True)


class SegmentZarrModule(argschema.ArgSchemaParser):
    default_schema = SegmentZarrParameters

    def output(self, d):
        out_json = self.args.get("output_json")
        if out_json:
            pathlib.Path(out_json).parent.mkdir(parents=True, exist_ok=True)
            with open(out_json, "w") as f:
                json.dump(f, d)
    
    @property
    def predict_options(self):
        if self.args["predict_options"]==None:
            return {'gpu_device':None,'bound_box':None}
        else:
            return self.args["predict_options"]

    def run(self):
        run(self.args["weights_file"], self.args["input_zarr"],
            self.args["probability_output"],
            self.args["zarr_level"],
            self.args["filter_max_intensity"],
            self.predict_options)
        self.output(self.args)


if __name__ == "__main__":
    mod = SegmentZarrModule()
    mod.run()

__all__ = [
    "SegmentZarrModule",
    "SegmentZarrParameters"]



import pandas as pd
import numpy as np
import zarr
import docker
import tensorstore as ts
import os
import json
import boto3

def create_tensor(fpath, arr_shape, driver='zarr', store='file', dtype='float32', fill_value=-np.inf, 
                       chunk_shape=[64, 64, 64], res=[1,1,1], scale=0, arr=None, AWS_Key=None, AWS_Secret_Key=None):
    """Create a tensorstore object, with optional setting of array
       driver: Type of file, including zarr, n5, precomputed
       store: Type of source, including file, in-memory, s3
       AWS Key, AWS_Secret_Key: Only applicable to s3 store
    """
    if 'int' in str(dtype):
        fill_value=0
    if isinstance(arr, np.ndarray):
        arr = arr.astype(dtype)
    kvstore = {"driver": store,"path": fpath}
    if store == 's3':
        if not AWS_Key or not  AWS_Secret_Key:
            raise TypeError("AWS_Key and AWS_Secret_Key required for the S3 store")
        os.environ['AWS_ACCESS_KEY_ID'], os.environ['AWS_SECRET_ACCESS_KEY']=AWS_Key, AWS_Secret_Key
        bucket = fpath.split("/")[0]
        path = fpath.replace(bucket+"/", '')
        kvstore = {"driver": "s3","bucket": bucket ,"path": path}

    if driver in ['zarr','n5']:
        fill_value=None if driver=='n5' else fill_value
        out_arr = ts.open({
         'driver': driver,
         'kvstore': kvstore,
         },
         dtype=dtype,
         fill_value=fill_value,
         chunk_layout=ts.ChunkLayout(chunk_shape=chunk_shape),
         create=True,
         shape=list(arr_shape)).result()

    if driver == 'neuroglancer_precomputed':
        arr_shape=list(arr_shape)+[1] if len(arr_shape)==3 else arr_shape
        out_arr = ts.open(
                    {
                        "driver": "neuroglancer_precomputed",
                        "kvstore": kvstore,
                        "scale_metadata": {
                            "resolution": res,
                            "chunk_size": list(chunk_shape),
                            "encoding": "raw",
                            "key": "s" + str(scale)
                        }
                    },
                    create=True,
                    dtype=dtype,
                    domain=ts.IndexDomain(
                        shape=list(list(arr_shape)),
                    )).result()

    if isinstance(arr, np.ndarray):
        out_arr.write(arr).result()
        
    os.environ['AWS_ACCESS_KEY_ID'], os.environ['AWS_SECRET_ACCESS_KEY'] = '', ''
    return out_arr

def open_tensor(fpath, driver='zarr', store='file', AWS_Key=None, AWS_Secret_Key=None, bytes_limit= 100_000_000):
    """Open a tensorstore object.
       driver: Type of file, including zarr, n5, precomputed
       store: Type of source, including file, s3
       AWS Key, AWS_Secret_Key: Only applicable to s3 store
    """

    kvstore = {"driver": store,"path": fpath}
    if store == 's3':
        if not AWS_Key or not  AWS_Secret_Key:
            raise TypeError("AWS_Key and AWS_Secret_Key required for the S3 store")
        os.environ['AWS_ACCESS_KEY_ID'], os.environ['AWS_SECRET_ACCESS_KEY']=AWS_Key, AWS_Secret_Key
        bucket = fpath.split("/")[0]
        path = fpath.replace(bucket+"/", '')
        kvstore = {"driver": "s3","bucket": bucket ,"path": path}
    #Load tensorstore array
    dataset_future = ts.open({
         'driver':
             driver,
         'kvstore': kvstore,
     # Use 100MB in-memory cache.
         'context': {
             'cache_pool': {
                 'total_bytes_limit': bytes_limit
             }
         },
         'recheck_cached_data':
         'open',
     })

    os.environ['AWS_ACCESS_KEY_ID'], os.environ['AWS_SECRET_ACCESS_KEY'] = '', ''
    return dataset_future.result()

create_EmptyTensor = create_tensor  
open_ZarrTensor = open_tensor
    
    
def zarr_to_n5(zarr_path, out_path, chunks=(64,64,64), cutout=None):
    #open zarr
    arr = open_ZarrTensor(zarr_path)
    if cutout != None:
        x1,x2,y1,y2,z1,z2 = cutout
        arr = arr[0,0,x1:x2,y1:y2,z1:z2].transpose().read().result()
    else:
        arr = arr[0,0,:,:,:].transpose().read().result()

    #create n5
    store = zarr.N5Store(os.path.join(out_path, '.n5'))
    root = zarr.group(store=store)
    z = root.zeros('group/' + zarr_path[-2], shape=arr.shape, chunks=chunks, dtype=arr.dtype, compressor=None)
    z[:] = arr
    
def zarr_to_precomputed(zarr_path, out_path, store='file', chunks=(64,64,64), cutout=None, scales=6, AWS_Key=None, AWS_Secret_Key=None):
    #iterate over all scale levels
    for scale in range(0,scales):
        #open zarr
        arr = open_tensor(zarr_path+str(scale))
        if cutout!=None:
            x1,x2,y1,y2,z1,z2 = cutout
            arr = arr[0,0,x1:x2,y1:y2,z1:z2].read().result()
            cutout = list((np.array(cutout)/2).astype(int))
        else:
            arr = arr[0,0,:,:,:].read().result()
        arr = np.expand_dims(arr, axis=3)
    
        #get resolution
        r_path = os.path.join(os.path.dirname(zarr_path), ".zattrs")
        res = json.loads(open(r_path, "r").read())['multiscales'][0]['datasets'][int(scale)]['coordinateTransformations'][0]['scale'][2:]
        
        #create precomputed tensor
        pre_comp = create_tensor(out_path, arr_shape=arr.shape, dtype=arr.dtype, store=store, driver='neuroglancer_precomputed', 
                                      AWS_Key=AWS_Key, AWS_Secret_Key=AWS_Secret_Key, scale=scale)
        pre_comp.write(arr).result()
        

def zarr_to_CATMAID_project(zarr_path, out_path, container_id, AWS_Key, AWS_Secret_Key, project='NewProject', 
                            stack='NewStack', store='s3', chunks=(64,64,64), translation=(0,0,0), cutout=None, tile_dim=[128,128], ortho=False):

    #convert zarr to precomputed
    zarr_to_precomputed(zarr_path=zarr_path, AWS_Key=AWS_Key, AWS_Secret_Key=AWS_Secret_Key, 
                           out_path=out_path, store=store, chunks=(64,64,64), cutout=cutout)
    
    #extract url 
    split = out_path.split("/")
    bucket = split[0]
    key = out_path.replace(bucket+"/", '')
    url = os.path.join("https://", bucket+".s3-us-west-2.amazonaws.com", key, "%SCALE_DATASET%/")
    
    #get resolution and shape
    shape = open_tensor(zarr_path + "0/")[0,0,:,:,:].shape
    if cutout!=None:
        x1,x2,y1,y2,z1,z2 = cutout
        x,y,z = x2-x1,y2-y1,z2-z1
        shape = [x,y,z]
    r_path = os.path.join(os.path.dirname(os.path.dirname(zarr_path + "0/")), ".zattrs")
    res = json.loads(open(r_path, "r").read())['multiscales'][0]['datasets'][0]['coordinateTransformations'][0]['scale'][2:]
    
    #create project data json
    if ortho==True:
      stack_file = [{
        "project": {
          "title": project,
          "stacks": [{
            "title": stack+"xy",
            "dimension": str(tuple(shape)),
            "mirrors": [{
              "fileextension": "raw",
              "position": 0,
              "tile_source_type": 14,
              "tile_height":tile_dim[0],
              "tile_width":tile_dim[1],
              "title": stack+"_tiles",
              "url": url + "0_1_2"
            }],
            "resolution": str(tuple(res)),
            "translation": str(translation),
            "downsample_factors": ["(1,1,1)", "(2,2,2)", "(4,4,4)"],
            "orientation" : 0,
            "stackgroups": [{"title": project+"_StackGroup", "relation": "channel"}]
          },
          {
            "title": stack+"xz",
            "dimension": str(tuple([shape[0],shape[2],shape[1]])),
            "mirrors": [{
              "fileextension": "raw",
              "position": 0,
              "tile_source_type": 14,
              "tile_height":tile_dim[0],
              "tile_width":tile_dim[1],
              "title": stack+"_tiles",
              "url": url + "0_2_1"
            }],
            "resolution": str(tuple([res[0],res[2],res[1]])),
            "translation": str(tuple([translation[0],translation[2],translation[1]])),
            "downsample_factors": ["(1,1,1)", "(2,2,2)", "(4,4,4)"],
            "orientation" : 1,
            "stackgroups": [{"title": project+"_StackGroup", "relation": "view"}]
          },
          {
            "title": stack+"zy",
            "dimension": str(tuple([shape[2],shape[1],shape[0]])),
            "mirrors": [{
              "fileextension": "raw",
              "position": 0,
              "tile_source_type": 14,
              "tile_height":tile_dim[0],
              "tile_width":tile_dim[1],
              "title": stack+"_tiles",
              "url": url + "2_1_0"
            }],
            "resolution": str(tuple([res[2],res[1],res[0]])),
            "translation": str(tuple([translation[2],translation[1],translation[0]])),
            "downsample_factors": ["(1,1,1)", "(2,2,2)", "(4,4,4)"],
            "orientation" : 2,
            "stackgroups": [{"title": project+"_StackGroup", "relation": "view"}]
          }]
        }
      }]
      
    else:
      stack_file = [{
        "project": {
          "title": project,
          "stacks": [{
            "title": stack+"xy",
            "dimension": str(tuple(shape)),
            "mirrors": [{
              "fileextension": "raw",
              "position": 0,
              "tile_source_type": 14,
              "tile_height":tile_dim[0],
              "tile_width":tile_dim[1],
              "title": stack+"_tiles",
              "url": url + "0_1_2"
            }],
            "resolution": str(tuple(res)),
            "translation": str(translation),
            "downsample_factors": ["(1,1,1)", "(2,2,2)", "(4,4,4)"],
            "orientation" : 0,
            "stackgroups": [{"title": project+"_StackGroup", "relation": "channel"}]
          }]
        }
      }]
                                                              
    #save json to local                    
    json_fpath = os.path.join("./", "data.json")       
    with open(json_fpath, 'w') as f:
        json.dump(stack_file, f)
    
    #open docker container and import json
    client = docker.from_env()
    container = client.containers.get(container_id)
    copy_string = json_fpath + " " + container_id + ":" + "/home/django/projects/data.json"
    os.system("docker cp " + str(copy_string))
    container.exec_run('python3 manage.py catmaid_import_projects --input data.json --permission user:admin:can_import user:admin:can_annotate')

    #clean-up
    container.exec_run('rm input data.json')
    client.close()
    os.remove(json_fpath)

    return stack_file